"""Tool for automatically drawing images."""
from __future__ import annotations

import asyncio
import logging
from typing import Iterator

from PIL import Image

from .canvas import Canvas, Pixel
from .client import Client


logger = logging.getLogger('dpypx')


class AutoDrawer:
    """Tool for automatically drawing images."""

    @classmethod
    def load_image(
            cls, client: Client, xy: tuple[int, int],
            image: Image.Image, scale: float = 1) -> AutoDrawer:
        """Draw from the pixels of an image."""
        if image.mode == 'RGBA':
            new_image = Image.new('RGB', image.size)
            new_image.paste(image, mask=image)
            image = new_image
        width = round(image.width * scale)
        height = round(image.height * scale)
        # Since the image is only loaded once we can afford to use a high
        # quality filter.
        resized = image.resize((width, height), Image.LANCZOS)
        data = list(resized.getdata())
        grid = [
            [Pixel(*pixel) for pixel in data[start:start + width]]
            for start in range(0, len(data), width)
        ]
        return cls(client, *xy, grid)

    @classmethod
    def load(cls, client: Client, data: str) -> AutoDrawer:
        """Draw from a string that specifies the pixels.

        `data` should be a multi-line string. The first two lines are the x
        and y coordinates of the top left of the image to draw. The second two
        are the width and height of the image. The rest of the lines are the
        pixels of the image, as hex codes (horizontal scanlines, left-to-right
        top-to-bottom).
        """
        lines = data.split('\n')
        x = int(lines.pop(0))
        y = int(lines.pop(0))
        width = int(lines.pop(0))
        height = int(lines.pop(0))
        grid = []
        for _ in range(height):
            row = []
            for _ in range(width):
                row.append(Pixel.from_hex(lines.pop(0)))
            grid.append(row)
        return cls(client, x, y, grid)

    def __init__(
            self, client: Client, x: int, y: int, grid: list[list[Pixel]]):
        """Store the plan."""
        self.client = client
        self.grid = grid
        # Top left coords.
        self.x0 = x
        self.y0 = y
        # Bottom right coords.
        self.x1 = x + len(self.grid[0])
        self.y1 = y + len(self.grid)

    def _iter_coords(self) -> Iterator[tuple[int, int]]:
        """Iterate over the coordinates of the image."""
        for x in range(self.x0, self.x1):
            for y in range(self.y0, self.y1):
                yield x, y

    async def check_pixel(self, canvas: Canvas, x: int, y: int) -> bool:
        """Draw a pixel if not already drawn.

        Returns True if the pixel was not already drawn.
        """
        colour = self.grid[y - self.y0][x - self.x0]
        if canvas[x, y] == colour:
            logger.debug(f'Skipping already correct pixel at {x}, {y}.')
            return False
        await self.client.put_pixel(x, y, colour)
        return True

    async def draw(self):
        """Draw the pixels of the image, attempting each pixel max. once."""
        canvas = await self.client.get_canvas()
        for x, y in self._iter_coords():
            if await self.check_pixel(canvas, x, y):
                canvas = await self.client.get_canvas()

    async def draw_and_fix(self, forever: bool = True):
        """Draw the pixels of the image, prioritise fixing existing ones."""
        work_to_do = True
        while work_to_do:
            canvas = await self.client.get_canvas()
            work_to_do = False
            for x, y in self._iter_coords():
                if await self.check_pixel(canvas, x, y):
                    work_to_do = True
                    canvas = await self.client.get_canvas()
                    break
            if forever and not work_to_do:
                logger.info('Entire image is correct, waiting 1s to loop.')
                await asyncio.sleep(1)
                work_to_do = True



"""HTTP client to the pixel API."""
from __future__ import annotations

import logging
from typing import Union, Optional

import aiohttp

from .canvas import Canvas, Pixel
from .colours import Colour, parse_colour
from .errors import (
    HttpClientError, MethodNotAllowedError, RatelimitedError, ServerError
)
from . import ratelimits


logger = logging.getLogger('dpypx')


class Client:
    """HTTP client to the pixel API."""

    def __init__(
            self,
            token: str,
            base_url: str = 'https://pixels.pythondiscord.com/',
            *,
            # Param exists for backwards compatibility, no longer needed.
            ratelimit_save_file: Optional[str] = None):
        """Store the token and set up the client."""
        self.base_url = base_url
        self.headers = {
            'Authorization': 'Bearer ' + token,
            'User-Agent': 'Artemis dpypx (Python/aiohttp)'
        }
        self.client = None
        self.ratelimits = ratelimits.RateLimiter(self)

    async def get_client(self) -> aiohttp.ClientSession:
        """Get or create the client session."""
        if (not self.client) or self.client.closed:
            self.client = aiohttp.ClientSession(headers=self.headers)
        return self.client

    async def send_request(
            self,
            method: str,
            endpoint: str,
            data: Optional[dict] = None,
            params: Optional[dict] = None) -> Union[dict, bytes, None]:
        """Make a basic HTTP request to the API."""
        logger.debug(
            f'Request: {method} {endpoint} data={data!r} params={params!r}.'
        )
        client = await self.get_client()
        request = client.request(
            method, self.base_url + endpoint, json=data, params=params
        )
        async with request as response:
            if 500 > response.status >= 400:
                if response.status == 405:
                    class_ = MethodNotAllowedError
                elif response.status == 429:
                    class_ = RatelimitedError
                else:
                    class_ = HttpClientError
                if method == 'HEAD':
                    detail = 'No body (HEAD request).'
                else:
                    data = await response.json()
                    detail = data.get('message', data.get(
                        'detail', 'No error message.'
                    ))
                raise class_(response.status, detail)
            if response.status >= 500:
                raise ServerError()
            self.ratelimits.update(endpoint, response.headers)
            if method == 'HEAD':
                return
            elif response.headers['Content-Type'] == 'application/json':
                return await response.json()
            else:
                return await response.read()

    async def request(
            self,
            method: str,
            endpoint: str,
            *,
            data: Optional[dict] = None,
            params: Optional[dict] = None,
            ratelimit_after: bool = False) -> Union[dict, bytes, None]:
        """Make a call to an endpoint, respecting ratelimiting."""
        retry = True
        while retry:
            # Always check before sending a request, even if we *want* to wait
            # after, sending a request and getting ratelimited is undesirable.
            await self.ratelimits.pause(endpoint)
            try:
                resp = await self.send_request(method, endpoint, data, params)
            except RatelimitedError:
                retry = True
            else:
                retry = False
            if ratelimit_after:
                await self.ratelimits.pause(endpoint)
        return resp

    async def put_pixel(
            self, x: int, y: int, colour: Union[int, str, Colour]) -> str:
        """Draw a pixel and return a message."""
        # Wait for ratelimits *after* making request, not before. This makes
        # sense because we don't know how the canvas may have changed by the
        # time we have finished waiting, whereas for GET endpoints, we want to
        # return the information as soon as it is given.
        data = await self.request('POST', 'set_pixel', data={
            'x': x,
            'y': y,
            'rgb': parse_colour(colour)
        }, ratelimit_after=True)
        logger.info('Success: {message}'.format(**data))
        return data['message']

    async def get_canvas_size(self) -> tuple[int, int]:
        """Get the size of the canvas."""
        data = await self.request('GET', 'get_size')
        return data['width'], data['height']

    async def get_canvas(self) -> Canvas:
        """Request the entire canvas."""
        data = await self.request('GET', 'get_pixels')
        size = await self.get_canvas_size()
        return Canvas(size, data)

    async def get_pixel(self, x: int, y: int) -> Pixel:
        """Get a specific pixel of the canvas."""
        data = await self.request('GET', 'get_pixel', params={'x': x, 'y': y})
        return Pixel.from_hex(data['rgb'])

    async def swap_pixels(
            self, xy0: tuple[int, int], xy1: tuple[int, int]) -> str:
        """Swap two pixels on the canvas."""
        data = await self.request('POST', 'swap_pixel', data={
            'origin': {
                'x': xy0[0],
                'y': xy0[1]
            },
            'dest': {
                'x': xy1[0],
                'y': xy1[1]
            }
        }, ratelimit_after=True)
        logger.info('Success: {message}'.format(**data))
        return data['message']

    async def close(self):
        """Close the underlying session."""
        await self.client.close()



"""Utility for obeying ratelimits."""
from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass
from typing import Optional

from . import client
from .errors import MethodNotAllowedError


logger = logging.getLogger('dpypx')


@dataclass
class RateLimitEndpoint:
    """Ratelimiter for a specific endpoint."""

    client: client.Client
    endpoint: str
    # We don't care about Requests-Period, we just dynamically wait based on
    # Requests-Reset.
    ratelimited: Optional[bool] = None
    limit: Optional[int] = None
    remaining: Optional[int] = None
    reset: Optional[float] = None
    cooldown_reset: Optional[int] = None

    def update(self, headers: dict[str, int]):
        """Update the ratelimiter based on the latest headers."""
        if 'Cooldown-Reset' in headers:
            self.remaining = 0
            self.cooldown_reset = int(headers['Cooldown-Reset'])
            return
        if 'Requests-Remaining' not in headers:
            self.ratelimited = False
            return
        self.ratelimited = True
        self.remaining = int(headers['Requests-Remaining'])
        self.limit = int(headers['Requests-Limit'])
        self.reset = float(headers['Requests-Reset'])

    async def pause(self):
        """Pause before sending another request if necessary."""
        if self.cooldown_reset:
            logger.warning(f'Cooldown: Sleeping for {self.cooldown_reset}s.')
            await asyncio.sleep(self.cooldown_reset)
            self.cooldown_reset = None
            return
        if self.ratelimited is None:
            await self.check_limits()
        if not self.ratelimited:
            return
        if self.remaining:
            logger.debug(
                f'Not sleeping, {self.remaining} remaining requests.'
            )
            return
        if self.reset:
            logger.warning(f'Sleeping for {self.reset}s.')
            await asyncio.sleep(self.reset)

    async def check_limits(self):
        """Check the ratelimits with a HEAD request."""
        try:
            # Client.send_request will call self.update.
            await self.client.send_request('HEAD', self.endpoint)
        except MethodNotAllowedError:
            self.ratelimited = False


class RateLimiter:
    """Ratelimiters for all the endpoints."""

    def __init__(self, client: client.Client):
        """Set up the ratelimiter."""
        self.client = client
        self.ratelimits = {}

    def __getitem__(self, endpoint: str) -> RateLimitEndpoint:
        """Get the ratelimiter for a specific endpoint."""
        if endpoint not in self.ratelimits:
            self.ratelimits[endpoint] = RateLimitEndpoint(
                self.client, endpoint
            )
        return self.ratelimits[endpoint]

    def update(self, endpoint: str, headers: dict[str, int]):
        """Update the ratelimits for an endpoint with the latest headers."""
        self[endpoint].update(headers)

    async def pause(self, endpoint: str):
        """Pause before sending another request for an endpoint."""
        await self[endpoint].pause()



"""Client to Python Discord Pixels."""
from .autodraw import AutoDrawer
from .canvas import Canvas, Pixel
from .client import Client
from .colours import Colour, Colour as Colours
from .errors import *



"""Collection of errors the wrapper can raise."""


__all__ = [
    'DpypxError',
    'HttpClientError',
    'RatelimitedError',
    'MethodNotAllowedError',
    'ServerError',
    'CanvasFormatError'
]


class DpypxError(Exception):
    """Base class for DPYPX errors."""


class HttpClientError(DpypxError):
    """Exception raised when the API returns a 400 range code."""

    def __init__(self, code: int, message: str):
        """Store the code and message."""
        self.code = code
        self.message = message
        super().__init__(f'Error {code} - {message}')


class RatelimitedError(HttpClientError):
    """Exception raised when the client is ratelimited (code 429).

    This should be caught by the client so consumers of the library should
    not need to handle this.
    """


class MethodNotAllowedError(HttpClientError):
    """Exception raised when the client sends the wrong HTTP method.

    This could mean that the client is buggy, but it is also raised sometimes
    during normal operation of the client (and is caught).
    """


class ServerError(DpypxError):
    """Exception raised when the API returns an 500 range code."""


class CanvasFormatError(DpypxError):
    """Exception raised when the canvas is badly formatted."""
